@@ -25,7 +25,7 @@ import pytest
 import numpy as np
 import nibabel as nb
 
-from ..utils import Flatten, ConvertWarp, Deoblique, Reoblique
+from ..utils import Flatten, ConvertWarp, Deoblique, Reoblique, PadSlices
 
 
 def test_Flatten(tmpdir):
@@ -106,3 +106,36 @@ def test_Xeoblique(tmpdir, angles, oblique):
     )
 
     assert np.allclose(nb.load(reoblique.out_epi).affine, affine)
+
+
+@pytest.mark.parametrize("in_shape,expected_shape,padded", [
+    ((2,2,2), (2,2,2), False),
+    ((2,2,3), (2,2,4), True),
+    ((3,3,2,2), (3,3,2,2), False),
+    ((3,3,3,2), (3,3,4,2), True),
+])
+def test_pad_slices(tmpdir, in_shape, expected_shape, padded):
+    tmpdir.chdir()
+
+    data = np.random.rand(*in_shape)
+    aff = np.eye(4)
+
+    # RAS
+    img = nb.Nifti1Image(data, aff)
+    img.to_filename("epi-ras.nii.gz")
+    res = PadSlices(in_file="epi-ras.nii.gz").run().outputs
+
+    # LPS
+    newaff = aff.copy()
+    newaff[0, 0] *= -1.0
+    newaff[1, 1] *= -1.0
+    newaff[:2, 3] = aff.dot(np.hstack((np.array(img.shape[:3]) - 1, 1.0)))[:2]
+    img2 = nb.Nifti1Image(np.flip(np.flip(data, 0), 1), newaff)
+    img2.to_filename("epi-lps.nii.gz")
+    res2 = PadSlices(in_file="epi-lps.nii.gz").run().outputs
+
+
+    out_ras = nb.load(res.out_file)
+    out_lps = nb.load(res2.out_file)
+    assert out_ras.shape == out_lps.shape == expected_shape
+    assert res.padded == padded
