@@ -29,14 +29,22 @@ from scipy import ndimage as ndi
 from scipy.sparse import vstack as sparse_vstack, csr_matrix, kron
 
 import nibabel as nb
+import nitransforms as nt
+from nitransforms.base import _as_homogeneous
 from bids.utils import listify
 
 
+def _clear_shifts(instance, attribute, value):
+    instance.shifts = None
+    return value
+
+
 @attr.s(slots=True)
 class B0FieldTransform:
     """Represents and applies the transform to correct for susceptibility distortions."""
 
     coeffs = attr.ib(default=None)
+    xfm = attr.ib(default=nt.linear.Affine(), on_setattr=_clear_shifts)
     shifts = attr.ib(default=None, init=False)
 
     def fit(self, spatialimage):
@@ -46,6 +54,12 @@ class B0FieldTransform:
         Implements Eq. :math:`\eqref{eq:1}`, interpolating :math:`f(\mathbf{s})`
         for all voxels in the target-image's extent.
 
+        Returns
+        -------
+        updated : :obj:`bool`
+            ``True`` if the internal field representation was fit,
+            ``False`` if cache was valid and will be reused.
+
         """
         # Calculate the physical coordinates of target grid
         if isinstance(spatialimage, (str, bytes, Path)):
@@ -58,14 +72,18 @@ class B0FieldTransform:
             if np.all(newshape == self.shifts.shape) and np.allclose(
                 newaff, self.shifts.affine
             ):
-                return
+                return False
 
         weights = []
         coeffs = []
 
         # Generate tensor-product B-Spline weights
         for level in listify(self.coeffs):
-            wmat = grid_bspline_weights(spatialimage, level)
+            self.xfm.reference = spatialimage
+            moved_cs = level.__class__(
+                level.dataobj, self.xfm.matrix @ level.affine, level.header
+            )
+            wmat = grid_bspline_weights(spatialimage, moved_cs)
             weights.append(wmat)
             coeffs.append(level.get_fdata(dtype="float32").reshape(-1))
 
@@ -77,6 +95,7 @@ class B0FieldTransform:
 
         # Cache
         self.shifts = nb.Nifti1Image(vsm, spatialimage.affine, None)
+        return True
 
     def apply(
         self,
@@ -124,6 +143,9 @@ class B0FieldTransform:
 
         """
         # Ensure the vsm has been computed
+        if isinstance(spatialimage, (str, bytes, Path)):
+            spatialimage = nb.load(spatialimage)
+
         self.fit(spatialimage)
         vsm = self.shifts.get_fdata().copy()
 
@@ -135,9 +157,21 @@ class B0FieldTransform:
         pe_axis = "ijk".index(pe_dir[0])
 
         # Map voxel coordinates applying the VSM
-        ijk_axis = tuple([np.arange(s) for s in vsm.shape])
-        voxcoords = np.array(np.meshgrid(*ijk_axis, indexing="ij"), dtype="float32")
-        voxcoords[pe_axis, ...] += vsm * ro_time
+        if self.xfm is None:
+            ijk_axis = tuple([np.arange(s) for s in vsm.shape])
+            voxcoords = np.array(
+                np.meshgrid(*ijk_axis, indexing="ij"), dtype="float32"
+            ).reshape(3, -1)
+        else:
+            # Map coordinates from reference to time-step
+            hmc_xyz = self.xfm.map(self.xfm.reference.ndcoords.T)
+            # Convert from RAS to voxel coordinates
+            voxcoords = (
+                np.linalg.inv(self.xfm.reference.affine)
+                @ _as_homogeneous(np.vstack(hmc_xyz), dim=self.xfm.reference.ndim).T
+            )[:3, ...]
+
+        voxcoords[pe_axis, ...] += vsm.reshape(-1) * ro_time
 
         # Prepare data
         data = np.squeeze(np.asanyarray(spatialimage.dataobj))
@@ -146,7 +180,7 @@ class B0FieldTransform:
         # Resample
         resampled = ndi.map_coordinates(
             data,
-            voxcoords.reshape(3, -1),
+            voxcoords,
             output=output_dtype,
             order=order,
             mode=mode,
@@ -160,7 +194,7 @@ class B0FieldTransform:
         moved.header.set_data_dtype(output_dtype)
         return moved
 
-    def to_displacements(self, ro_time, pe_dir):
+    def to_displacements(self, ro_time, pe_dir, itk_format=True):
         """
         Generate a NIfTI file containing a displacements field transform compatible with ITK/ANTs.
 
@@ -181,38 +215,36 @@ class B0FieldTransform:
             A NIfTI 1.0 object containing the distortion.
 
         """
-        from math import pi
-        from nibabel.affines import voxel_sizes, obliquity
-        from nibabel.orientations import io_orientation
-
-        # Generate warp field
-        data = self.shifts.get_fdata(dtype="float32").copy()
+        # Set polarity & scale VSM (voxel-shift-map) by readout time
+        vsm = self.shifts.get_fdata().copy()
         pe_axis = "ijk".index(pe_dir[0])
-        pe_sign = -1.0 if pe_dir.endswith("-") else 1.0
-        pe_size = self.shifts.header.get_zooms()[pe_axis]
-        data *= pe_sign * ro_time * pe_size
+        vsm *= -1.0 if pe_dir.endswith("-") else 1.0
+        vsm *= ro_time
 
-        fieldshape = tuple(list(data.shape[:3]) + [3])
+        # Shape of displacements field
+        # Note that ITK NIfTI fields are 5D (have an empty 4th dimension)
+        fieldshape = tuple(list(vsm.shape[:3]) + [1, 3])
 
-        # Compose a vector field
-        field = np.zeros((data.size, 3), dtype="float32")
-        field[..., pe_axis] = data.reshape(-1)
+        # Convert VSM to voxel displacements
+        ijk_deltas = np.zeros((vsm.size, 3), dtype="float32")
+        ijk_deltas[:, pe_axis] = vsm.reshape(-1)
 
-        # If coordinate system is oblique, project displacements through directions matrix
-        aff = self.shifts.affine
-        if obliquity(aff).max() * 180 / pi > 0.01:
-            dirmat = np.eye(4)
-            dirmat[:3, :3] = aff[:3, :3] / (
-                voxel_sizes(aff) * io_orientation(aff)[:, 1]
-            )
-            field = nb.affines.apply_affine(dirmat, field)
+        # To convert from VSM to RAS field we just apply the affine
+        aff = self.shifts.affine.copy()
+        aff[:3, 3] = 0  # Translations MUST NOT be applied, though.
+        xyz_deltas = nb.affines.apply_affine(aff, ijk_deltas)
+        if itk_format:
+            # ITK displacement vectors are in LPS orientation
+            xyz_deltas[..., (0, 1)] *= -1.0
 
-        warpnii = nb.Nifti1Image(
-            field.reshape(fieldshape)[:, :, :, np.newaxis, :], aff, None
+        xyz_nii = nb.Nifti1Image(
+            xyz_deltas.reshape(fieldshape),
+            self.shifts.affine,
+            None,
         )
-        warpnii.header.set_intent("vector", (), "")
-        warpnii.header.set_xyzt_units("mm")
-        return warpnii
+        xyz_nii.header.set_intent("vector", (), "")
+        xyz_nii.header.set_xyzt_units("mm")
+        return xyz_nii
 
 
 def _cubic_bspline(d):
@@ -294,3 +326,14 @@ def grid_bspline_weights(target_nii, ctrl_nii):
         wd.append(csr_matrix(weights))
 
     return kron(kron(wd[0], wd[1]), wd[2])
+
+
+def _move_coeff(in_coeff, fmap_ref, transform):
+    """Read in a rigid transform from ANTs, and update the coefficients field affine."""
+    xfm = nt.linear.Affine(
+        nt.io.itk.ITKLinearTransform.from_filename(transform).to_ras(),
+        reference=fmap_ref,
+    )
+    coeff = nb.load(in_coeff)
+    newaff = xfm.matrix @ coeff.affine
+    return coeff.__class__(coeff.dataobj, newaff, coeff.header)
