@@ -108,23 +108,38 @@ def find_estimators(*, layout, subject, fmapless=True, force_fmapless=False):
      FieldmapEstimation(sources=<2 files>, method=<EstimatorType.PEPOLAR: 2>,
                         bids_id='auto_00007')]
 
-    Finally, *SDCFlows*' "*dataset A*" contains a BIDS structure with zero-bytes
-    NIfTI files and some corresponding metadata:
+    Finally, *SDCFlows*' "*dataset A*" and "*dataset B*" contain BIDS structures
+    with zero-byte NIfTI files and some corresponding metadata:
 
     >>> find_estimators(
     ...     layout=layouts['dsA'],
     ...     subject="01",
     ... )  # doctest: +ELLIPSIS
     [FieldmapEstimation(sources=<2 files>, method=<EstimatorType.MAPPED: 4>,
-                        bids_id='auto_00008'),
+                        bids_id='auto_...'),
      FieldmapEstimation(sources=<4 files>, method=<EstimatorType.PHASEDIFF: 3>,
-                        bids_id='auto_00009'),
+                        bids_id='auto_...'),
      FieldmapEstimation(sources=<3 files>, method=<EstimatorType.PHASEDIFF: 3>,
-                        bids_id='auto_00010'),
+                        bids_id='auto_...'),
      FieldmapEstimation(sources=<4 files>, method=<EstimatorType.PEPOLAR: 2>,
-                        bids_id='auto_00011'),
+                        bids_id='auto_...'),
      FieldmapEstimation(sources=<2 files>, method=<EstimatorType.PEPOLAR: 2>,
-                        bids_id='auto_00012')]
+                        bids_id='auto_...')]
+
+    >>> find_estimators(
+    ...     layout=layouts['dsB'],
+    ...     subject="01",
+    ... )  # doctest: +ELLIPSIS
+    [FieldmapEstimation(sources=<2 files>, method=<EstimatorType.MAPPED: 4>,
+                        bids_id='auto_...'),
+     FieldmapEstimation(sources=<4 files>, method=<EstimatorType.PHASEDIFF: 3>,
+                        bids_id='auto_...'),
+     FieldmapEstimation(sources=<3 files>, method=<EstimatorType.PHASEDIFF: 3>,
+                        bids_id='auto_...'),
+     FieldmapEstimation(sources=<4 files>, method=<EstimatorType.PEPOLAR: 2>,
+                        bids_id='auto_...'),
+     FieldmapEstimation(sources=<2 files>, method=<EstimatorType.PEPOLAR: 2>,
+                        bids_id='auto_...')]
 
     After cleaning the registry, we can see how the "*fieldmap-less*" estimation
     can be forced:
@@ -168,15 +183,15 @@ def find_estimators(*, layout, subject, fmapless=True, force_fmapless=False):
     ...     force_fmapless=True,
     ... )  # doctest: +ELLIPSIS
     [FieldmapEstimation(sources=<2 files>, method=<EstimatorType.MAPPED: 4>,
-                        bids_id='auto_00006'),
+                        bids_id='auto_...'),
      FieldmapEstimation(sources=<4 files>, method=<EstimatorType.PHASEDIFF: 3>,
-                        bids_id='auto_00007'),
+                        bids_id='auto_...'),
      FieldmapEstimation(sources=<3 files>, method=<EstimatorType.PHASEDIFF: 3>,
-                        bids_id='auto_00008'),
+                        bids_id='auto_...'),
      FieldmapEstimation(sources=<4 files>, method=<EstimatorType.PEPOLAR: 2>,
-                        bids_id='auto_00009'),
+                        bids_id='auto_...'),
      FieldmapEstimation(sources=<2 files>, method=<EstimatorType.PEPOLAR: 2>,
-                        bids_id='auto_00010')]
+                        bids_id='auto_...')]
 
     This function should also correctly investigate multi-session datasets:
 
@@ -209,6 +224,8 @@ def find_estimators(*, layout, subject, fmapless=True, force_fmapless=False):
         "scope": "raw",  # Ensure derivatives are not captured
     }
 
+    subject_root = Path(layout.root) / f"sub-{subject}"
+
     fmapless = fmapless or {}
     if fmapless is True:
         fmapless = {"bold", "dwi"}
@@ -252,7 +269,6 @@ def find_estimators(*, layout, subject, fmapless=True, force_fmapless=False):
         if epi_fmap.path in fm._estimators.sources:
             continue  # skip EPI images already considered above
 
-        subject_root = Path(epi_fmap.path.rpartition("/sub-")[0]).parent
         targets = [epi_fmap] + [
             layout.get_file(str(subject_root / intent))
             for intent in epi_fmap.get_metadata()["IntendedFor"]
@@ -279,10 +295,6 @@ def find_estimators(*, layout, subject, fmapless=True, force_fmapless=False):
 
     from .epimanip import get_trt
 
-    intended_root = Path(anat_file[0].path).parent.parent
-    if sessions:
-        intended_root = intended_root.parent
-
     for ses, suffix in sorted(product(sessions or (None,), fmapless)):
         candidates = layout.get(suffix=suffix, session=ses, **base_entities)
 
@@ -311,7 +323,7 @@ def find_estimators(*, layout, subject, fmapless=True, force_fmapless=False):
                 fmfiles, fmpaths = tuple(
                     zip(
                         *[
-                            (target, str(Path(target.path).relative_to(intended_root)))
+                            (target, str(Path(target.path).relative_to(subject_root)))
                             for i, target in enumerate(epi_targets)
                             if pe_dirs[i] == pe_dir and ro_totals[i] == ro_time
                         ]
